! =====================================================================
!     ****** CODICI/mm
!
!     COPYRIGHT
!       (c) 2002/2012 by CASPUR/G.Amati,F.Salvadore
!     NAME
!       mm
!     DESCRIPTION
!       Matrix Multiplication
!       naive version: no optimization
!     INPUTS
!       none
!     OUTPUT
!       none
!     TODO
!
!     NOTES
!       integer variables used: i,j,k,n
!       real variables used:    randd,a,b,c,mflop
!                               time1, time2
!
!     *****
! =====================================================================
!
program mm
!
    use precision_module
    use timing_module
       
    implicit none
#ifdef PAPI
#include "fpapi_test.h"
#endif
!
    integer i,j,k ! index
    integer, parameter :: n=512            ! size of the matrix
!
    real(my_kind) a(1:n,1:n) ! matrix
    real(my_kind) b(1:n,1:n) ! matrix
    real(my_kind) c(1:n,1:n) ! matrix (destination)
    real(my_kind) time1, time2 ! timing
    real(my_kind) mflops ! Mflops
#ifdef PAPI
      character*20  nome
      integer check,retval,event_code
      integer*8    :: values(100),event_set
      integer events(100)
#endif

#ifdef PAPI
          do i=1,100
            values=0
          end do

          check = PAPI_VER_CURRENT
          call PAPIF_library_init(check)
          if (check .ne. PAPI_VER_CURRENT) then
             print *, 'Error initializing the PAPI Library'
             call abort
          end if

#ifdef CACHE   
          events(1)=PAPI_L2_TCA
          events(2)=PAPI_L2_TCM
          events(3)=PAPI_L3_TCA
          events(4)=PAPI_L3_TCM

#else   
          events(1)=PAPI_DP_OPS
          events(2)=PAPI_TOT_CYC
#endif

              event_set = PAPI_NULL
          call PAPIF_create_eventset(event_set, check)
          if (check .ne. PAPI_OK) then
            print *, 'Error in subroutine PAPIF_create_eventset'
            call abort
          end if

#ifdef CACHE
         do i=1,4
#else
         do i=1,2
#endif
          event_code = events(i) 
          call PAPIf_query_event(event_code, retval)
            if (retval .NE. PAPI_OK) then
                write(*,*) 'non esiste ',values(i)
           end if
          call PAPIF_add_event(event_set, event_code, check)
          if (check .NE. PAPI_OK) then
             print *, "Abort After PAPIF_add_events: ",event_code, check
            call abort
          endif
        end do
#endif

!
    mflops = 2*float(n)*float(n)*float(n)/(1000.0*1000.0)
!
    call timing(time1)
    call random_number(a)
    call random_number(b)
    c = 0._my_kind
    call timing(time2)
    write(*,*) "--------------------------------------"
    write(*,*) " Matrix-Matrix Multiplication         "
    write(*,*) " Calcolo con precisione: ",precision(a(1,1))
    write(*,*) " rel. 0, naive multiplication         "
    write(*,*) " size =", n                      
    write(*,*) "--------------------------------------"
    write(*,*) "initialization", time2-time1
    write(*,*) a(n/2,n/2),b(n/2,n/2),c(n/2,n/2)
!
! main loop
!
    call timing(time1)
#ifdef PAPI
       call PAPIF_start(event_set, check)
#endif
!
!  qui sotto va completato il programma, prodotto righe per colonne
!   c_ij = c_ij + a_ik * b_kj     
!
    do j = 1,n
       do k = 1,n
          do i = 1,n
             c(i,j) = c(i,j) + a(i,k)*b(k,j)
          enddo
       enddo
    enddo
!
!  qui sopra va completato il programma, prodotto righe per colonne
!
#ifdef PAPI
      call PAPIF_stop(event_set, values, check)
#endif
    call timing(time2)
    write(*,*) "--------------------------------------"
    write(*,*) "time for moltiplication", time2-time1
    write(*,*) "Mflops                 ", mflops/(time2-time1)
    write(*,*) "--------------------------------------"
    write(*,*) "       "
    write(*,*) "       "
    write(*,*) a(n/2,n/2),b(n/2,n/2),c(n/2,n/2)
#ifdef PAPI
    write(*,*) "-------- OUTPUT PAPI  --------------------------"
#ifdef CACHE
    write(*,*) 'PAPI L2 %  ', 100*(float(values(2))/float(values(1)))
    write(*,*) 'PAPI L3 %  ', 100*(float(values(4))/float(values(3)))
    write(*,*) "--------------------------------------"
#else
    write(*,*) 'PAPI flops  ', values(1)/((time2-time1)*10**6)
    write(*,*) 'PAPI cicli  ', values(2)
    write(*,*) 'operazioni medie per ciclo  ', float(values(1))/float(values(2))
#endif

#endif
!
end program mm

